/*******************************************************************
  MX1-Loader - DragonBall M9328/MX1 loader

  mx1-load.c  - Universal MX1 loader main module

  (C) Copyright 2004 by Pavel Pisa - project originator
        http://cmp.felk.cvut.cz/~pisa
  (C) Copyright 2004 PiKRON Ltd.
        http://www.pikron.com
  
  USB support based on work of Roman Bartosinski (bartosr@centrum.cz)


  The MX1-Loader project can be used and distributed
  in compliance with any of next licenses
   - GPL - GNU Public License
     See file COPYING for details.
   - LGPL - Lesser GNU Public License
   - MPL - Mozilla Public License
   - and other licenses added by project originator

  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  WARNING: if you delete all lines, you are not allowed to
  distribute code or sources in any form.
 *******************************************************************/

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <bfd.h>
#include <inttypes.h>
#include "rs232_fn.h"

#include "universal-load.h"

char *mx1_hwlink="rs232-mx1";
char *mx1_sdev="/dev/ttyS0";
int mx1_baud=57600;
int mx1_waitrep=1000000;
int mx1_fd;
int mx1_xhz=18432000;

int go_flg=0;
int reset_flg=0;
int break_flg=0;
int upload_flg=0;
int regerase_flg=0;
int dump_flg=0;
int masserase_flg=0;
unsigned long masserase_end=0;
int fill_flg;
void *fill_pat_val;
int fill_pat_len;
char *file_format=NULL;

int command=0;
int blockmode=0;
int erase_block=-1;
int mem_type=0;
unsigned long mem_start=0;
unsigned long mem_length=0;
unsigned long mem_offs=0;
unsigned long go_addr=0;
unsigned long entry_addr=0;
uniload_alg_extprop_t *extprop_list=NULL;

int hashmark=1;
int uniload_load_use_lma=0;
static int load_section_error;

void
uniload_propeller(u_long addr, FILE * fp)
{
static char *str = "\\|/-";
static int index;
int tv_diff;
static struct timeval tv_old;
struct timeval tv_actual;
	
	gettimeofday(&tv_actual,NULL);
	tv_diff=tv_actual.tv_sec-tv_old.tv_sec;
	if(!tv_diff) return;
	
	tv_old=tv_actual;

	if (!hashmark) return;
	fprintf(fp, "%c 0x%08lx\b\b\b\b\b\b\b\b\b\b\b\b", str[index++], addr);
	fflush(fp);
	index %= 4;
}

#define MEM_DUMP_ATONCE 1024
#define MEM_MAX_ATONCE (1024*1024)

int uniload_mem_dump(uniload_alg_info_t *algi, int mem_type, unsigned long start, unsigned long len)
{
  int res;
  unsigned long addr=start;
  unsigned char buff[MEM_DUMP_ATONCE];
  unsigned char *p;
  int blen=MEM_DUMP_ATONCE;
  int cnt;
  int i;
  
  while(len){
    cnt=(len>blen)?blen:len;
    res=uniload_mem_read(algi, mem_type, addr, cnt, buff);
    if(res<0){
      fprintf(stderr,"uniload_mem_dump: uniload_mem_read returned %d\n", res);
      return -1;
    }
    len-=cnt;
    p=buff;
    while(cnt){
      printf("%08lX:",addr);
      i=cnt>16?16:cnt;
      addr+=i;
      cnt-=i;
      while(i--){
        printf("%02X%c",*(p++),i?' ':'\n');
      }
    }
  }
  return 0;  
}


int uniload_upload_file(uniload_alg_info_t *algi, char *file_name, char *format,
                        int mem_type, unsigned long start, unsigned long len)
{
  FILE *file;
  char *mode="wb";
  int res;
  unsigned long addr=start;
  unsigned char buff[MEM_DUMP_ATONCE];
  int blen=MEM_DUMP_ATONCE;
  int cnt;

  if(!strcmp(file_name,"-")) file_name=NULL;
  
  if(file_name){
    if((file=fopen(file_name,mode))==NULL)
    {perror("uniload_upload_file : file open");
     return -1;
    }
  } else file=stdout;

  while(len){
    cnt=(len>blen)?blen:len;
    res=uniload_mem_read(algi, mem_type, addr, cnt, buff);
    if(res<0){
      fprintf(stderr,"uniload_upload_file: uniload_mem_read returned %d\n", res);
      if(file_name) fclose(file); 
      return -1;
    }
    if(fwrite(buff,cnt,1,file)!=1){
      perror("upload_file : file write");
      return -1;
    }

    len-=cnt;
    addr+=cnt;
  }

  if(file_name) fclose(file); 

  return 0;
}

int uniload_mem_fill(uniload_alg_info_t *algi, int mem_type, unsigned long start, unsigned long len,
		unsigned char *pat_val, int pat_len)
{
  int res;
  unsigned long addr=start;
  unsigned char buff[MEM_DUMP_ATONCE];
  int blen=MEM_DUMP_ATONCE;
  unsigned char *p;
  int cnt;
  int i;
  
  if(pat_len<=0) return -1;
  
  if((pat_len<=blen/2)&&(len>pat_len)){
    i=blen<len+pat_len?blen:len+pat_len-1;
    if((pat_len==1)){
      blen=i;
      memset(buff,*pat_val,blen);
    }else{
      i=i/pat_len;
      blen=i*pat_len;
      for(p=buff;i;i--,p+=pat_len){
        memcpy(p,pat_val,pat_len);
      }
    }
    pat_val=buff;
  } else {
    blen=pat_len;
  }
  
  while(len){
    cnt=(len>blen)?blen:len;
    res=uniload_mem_write(algi, mem_type, addr, cnt, pat_val);
    if(res<0){
      fprintf(stderr,"uniload_mem_fill: uniload_mem_write returned %d\n", res);
      return -1;
    }
    len-=cnt;
    addr+=cnt;
  }
  return 0;  
}

int uniload_do_load_plain_binary(uniload_alg_info_t *algi, char *file_name)
{
  FILE *file;
  char *mode="rb";
  int res;
  unsigned long addr=algi->mem_offs;
  unsigned char *buff;
  int blen=MEM_DUMP_ATONCE;
  int cnt;

  if(!strcmp(file_name,"-")) file_name=NULL;
  
  if(file_name){
    if((file=fopen(file_name,mode))==NULL)
    {perror("uniload_do_load_plain_binary : file open");
     return -1;
    }
  } else file=stdin;

  if(algi->max_blksize){
    blen=algi->max_blksize<MEM_MAX_ATONCE?algi->max_blksize:MEM_MAX_ATONCE;
  }
  buff=malloc(blen);
  if(!buff){
    blen=MEM_DUMP_ATONCE;
    buff=malloc(blen);
    if(!buff)
      return -1;
  }

  while(1){
    cnt=fread(buff,1,blen,file);
    if(cnt==0)
      break;
    if(cnt<0){
      perror("uniload_do_load_plain_binary : file read");
      if(file_name) fclose(file);
      free(buff);
      return -1;
    }
      
    uniload_propeller(addr, stdout);
    res=uniload_mem_write(algi, algi->def_mem_type, addr, cnt, buff);
    if(res<0){
      fprintf(stderr,"uniload_do_load_plain_binary: uniload_mem_write addr 0x%lx returned %d\n",
      		 addr, res);
      if(file_name) fclose(file); 
      free(buff);
      return -1;
    }
    addr+=cnt;
  }

  if(file_name) fclose(file); 

  free(buff);
  return 0;
}


/*----------------------------------------------------------------*/

static void
uniload_load_section(bfd * abfd, sec_ptr sec, PTR priv)
{
	u_long addr;
	u_long load_bytes;
	file_ptr offset;
	int cnt;
	char *cbuf;
	uniload_alg_info_t *algi=(uniload_alg_info_t *)priv;
	int ret;
	long cbuf_size=algi->max_blksize;

	if(!cbuf_size)
	 	cbuf_size=512;

	dbprintf("uniload_load_section:\n\tsection %s index %d\n",
			sec->name, sec->index);
	dbprintf("\tflags %#x raw_size 0x%08lx cooked_size 0x%08lx\n",
			sec->flags, (long)sec->_raw_size, (long)sec->_cooked_size);
	dbprintf("\tvma %#lx lma %#lx output_offset %#lx\n",
			(long)sec->vma, (long)sec->lma, (long)sec->output_offset);
	if ((load_section_error < 0) ||
		((bfd_get_section_flags(abfd, sec) & SEC_LOAD) == 0)) {
		return;
	}
	if(!uniload_load_use_lma)
		addr = bfd_section_vma(abfd, sec);
	else
		addr = bfd_section_lma(abfd, sec);

	if ((load_bytes = bfd_section_size(abfd,sec)) == 0)
		return;

	if(cbuf_size>load_bytes)
		cbuf_size = load_bytes;
	
	cbuf = malloc(cbuf_size);
	
	if(!cbuf){
		dbprintf("cannot allocate cbuf %ld bytes\n", cbuf_size);
		load_section_error = -1;
		return;
	}

	offset = 0;
	while (load_bytes) {
		if (load_bytes > cbuf_size) {
			cnt = cbuf_size;
		} else {
			cnt = load_bytes;
		}

	        /* printf("cbuf_size %ld cnt %d\n",cbuf_size, cnt); */

		if (!bfd_get_section_contents(abfd, sec, cbuf, offset, cnt)) {
			dbprintf("read error section %s\n", sec->name);
			load_section_error = -1;
			free(cbuf);
			return;
		}
		/* mx1_wrb_filt(mx1_bfilt,(caddr_t) addr, cnt, cbuf)) */
		if ((ret=uniload_mem_write(algi, algi->def_mem_type, addr+algi->mem_offs, cnt, cbuf))<0) {
			dbprintf("write error on block write, written %d returned %d errno %d\n",
					cnt, ret, errno);
			load_section_error = -2;
			free(cbuf);
			return;
		}

		load_bytes -= cnt;
		addr += cnt;
		offset += cnt;
		uniload_propeller(addr, stdout);
	}
	free(cbuf);
}

static int bfd_initialized;
static bfd * prepare_binary(char *file_name)
{
	bfd *abfd;

	if (!bfd_initialized) {
		bfd_init();
		bfd_initialized = 1;
	}
	abfd = bfd_openr(file_name, 0);
	if (!abfd) {
		dbprintf("Unable to open file %s\n", file_name);
		return NULL;
	}
	if (bfd_check_format(abfd, bfd_object) == 0) {
		dbprintf("File %s is not an object file\n", file_name);
		bfd_close(abfd);
		return NULL;
	}
	return abfd;
}

/*
 * load a binary file, returns error codes
 */
int
uniload_do_load_binary(char *file_name, char *entry_name,
		u_long *entry_pt, uniload_alg_info_t *algi)
{
	bfd *abfd;
	u_long entry_addr = 0;

	if ((abfd = prepare_binary(file_name)) == NULL) 
		return -3;

	if (entry_name)
	    while (isspace(*entry_name))
		entry_name++;

	if (entry_name && isdigit(*entry_name)) {
	    entry_addr = strtoul (entry_name, NULL, 0);
	} else if (entry_name) {
	    long i, symcnt;
	    asymbol **symtab;
	    if ((i=bfd_get_symtab_upper_bound(abfd))<=0 || !(symtab=malloc(i)))
		return -3;
	    if ((symcnt = bfd_canonicalize_symtab (abfd, symtab))<0)
		return -3;

	    for (i=0; i<symcnt; i++) {
		if (!strcmp (entry_name, symtab[i]->name)) {
		    entry_addr = symtab[i]->section->vma + symtab[i]->value;
		    break;
		}
	    }
	} else {
	  entry_addr = bfd_get_start_address(abfd);
	}
	if(entry_pt) *entry_pt = entry_addr;

	load_section_error = 0;

	bfd_map_over_sections(abfd, uniload_load_section, algi);
	if (hashmark) {
		printf(" .          \b\b\b\b\b\b\b\b\b\b\b\b");
		fflush(stdout);
	}
	bfd_close(abfd);
	return load_section_error;
}

int
uniload_do_load_binary_section(char *file_name, char *section_name,
		uniload_alg_info_t *algi)
{
	bfd *abfd;
	asection *sec;

	if ((abfd = prepare_binary(file_name)) == NULL) 
		return -3;

	if ((sec = bfd_get_section_by_name(abfd, section_name)) == NULL)
		return -4;

	load_section_error = 0;

	if (hashmark) {
		printf("  "); fflush(stdout);
	}
	uniload_load_section(abfd, sec, algi);
	if (hashmark) {
		printf("\r         \r");
		fflush(stdout);
	}
	bfd_close(abfd);
	return load_section_error;
}


int uniload_download_file(uniload_alg_info_t *algi, char *file_name, 
		char *file_format, char *entry_name, u_long *entry_pt)
{
  int ret;
  
  if(file_format){
    if(!strcmp(file_format,"binary")){
      ret=uniload_do_load_plain_binary(algi, file_name);
      return ret;
    }
  }
 
  ret=uniload_do_load_binary(file_name, entry_name, entry_pt, algi);

  return ret;
}

/*----------------------------------------------------------------*/

int si_long(char **ps,long *val,int base)
{
  char *p;
  *val=strtol(*ps,&p,base);
  if(*ps==p) return -1;
  *ps=p;
  return 1;
}

int add_to_arr(void **pdata,int *plen,int base,char *str)
{
  char *s=str;
  long val;
  void *p;

  do{
    while(*s && strchr(", \t:;",*s)) s++;
    if(!*s) break;
    if(si_long(&s,&val,base)<0){
      return -1;
    }
    if(*pdata==NULL){
      *plen=0;
      *pdata=p=malloc(1);
    }else{
      p=realloc(*pdata,*plen+1);
      if(p==NULL) return -1;
      *pdata=p;
    }
    ((unsigned char*)p)[*plen]=val;
    (*plen)++;
  } while(1);
  return 1;
}

/*----------------------------------------------------------------*/

int uniload_extprop_add(uniload_alg_info_t *algi, uniload_alg_extprop_t *extprop)
{
  extprop->next=algi->extprop;
  algi->extprop=extprop;
  return 0;
}

uniload_alg_extprop_t *uniload_extprop_find(uniload_alg_info_t *algi, char *key)
{
  uniload_alg_extprop_t *extprop=algi->extprop;
  while(extprop){
    if(!strcmp(extprop->key,key))
      return extprop;
    extprop=extprop->next;
  }
  return NULL;
}

int uniload_extprop_get_ulong(uniload_alg_info_t *algi, char *key, unsigned long *pval)
{
  unsigned long val;
  char *p;
  uniload_alg_extprop_t *extprop;
  extprop=uniload_extprop_find(algi,key);
  if(!extprop)
    return 0;
  p=extprop->val;
  if(si_long(&p,&val,0)>0){
    *pval=val;
    return 1;
  }else{
    return -1;
  }
}

int extprop_parse2list(uniload_alg_extprop_t **extprop_list, char *option)
{
  uniload_alg_extprop_t *extprop;
  char *p;
  int key_len;

  p=strchr(option,'=');
  if(!p) return -1;

  key_len=p-option;
  extprop=malloc(sizeof(uniload_alg_extprop_t));
  if(!extprop)
    return -2;
  memset(extprop,0,sizeof(uniload_alg_extprop_t));
  extprop->key=malloc(key_len+1);
  if(!extprop->key){
    free(extprop);
    return -1;
  }
  strncpy(extprop->key,option,key_len);
  extprop->key[key_len]=0;

  extprop->val=p+1;

  extprop->next=*extprop_list;
  *extprop_list=extprop;
  return 1;
}

/*----------------------------------------------------------------*/

#ifndef __u8
  #define __u8 uint8_t
#endif
#ifndef __u16
  #define __u16 uint16_t
#endif
#ifndef __u32
  #define __u32 uint32_t
#endif

#include "linux_setup.h"

typedef struct atag_container {
  unsigned long base;
  uint8_t *data;
  unsigned long size;
  unsigned long capacity;
} atag_container_t;

typedef struct {
  char *name;
  uint32_t tag;
} atag_std_names_t;

atag_std_names_t atag_std_names[]={
  {"none"	, ATAG_NONE},
  {"core"	, ATAG_CORE},
  {"mem"	, ATAG_MEM},
  {"videotext"	, ATAG_VIDEOTEXT},
  {"ramdisk"	, ATAG_RAMDISK},
  {"initrd"	, ATAG_INITRD},
  {"initrd2"	, ATAG_INITRD2},
  {"serial"	, ATAG_SERIAL},
  {"revision"	, ATAG_REVISION},
  {"videolfb"	, ATAG_VIDEOLFB},
  {"cmdline"	, ATAG_CMDLINE},
  {"acorn"	, ATAG_ACORN},
  {"memclk"	, ATAG_MEMCLK},
  {"memdata"	, ATAG_MEMDATA},
  {"memcpy"	, ATAG_MEMCPY},
  {"jump"	, ATAG_JUMP},
  {NULL		, 0}
};

void *atag_add_space(atag_container_t *container, uint32_t bytes)
{
  uint32_t *p;
  bytes=(bytes+3)&~3;
  long size=container->size+bytes;
  long capacity;
  if(size>container->capacity){
    capacity=size+0x100;
    if(container->data){
      container->data=realloc(container->data,capacity);
    }else{
      container->data=malloc(capacity);
    }
    container->capacity=capacity;
  }
  if(!container->data)
    return NULL;
  p=(uint32_t*)(container->data+container->size);
  memset(p,0,bytes);
  container->size=size;
  return p;
}

void *atag_add_head(atag_container_t *container, uint32_t size, uint32_t tag)
{
  uint32_t *p;
  p=(uint32_t*)atag_add_space(container, size*4);
  if(!p) return NULL;
  p[0]=size;
  p[1]=tag;
  return p+2;
}

int atag_add_string_tag(atag_container_t *container, uint32_t tag, char *str)
{
  void *p;
  int l=strlen(str)+1+8;
  p=atag_add_head(container, (l+3)>>2, tag);
  if(!p) return 0;
  strcpy(p,str);
  return 1;
}

int atag_parse_tag(char **ps, uint32_t *tag)
{
  char *p, *r;
  int tlen;
  atag_std_names_t *a;
  long l;
  
  r=*ps;
  p=strchr(r,'=');
  if(!p) return -1;

  tlen=p-r;

  if(si_long(&r,&l,0)>0){
    if((*r!='=') && *r)
      return -1;
    *tag=l;
    *ps=p+1;
    return 1;
  }

  r=*ps;
  for(a=atag_std_names;a->name;a++){
    if(strlen(a->name)!=tlen) continue;
    if(memcmp(a->name,r,tlen)) continue;
    *tag=a->tag;
    *ps=p+1;
    return 1;
  }
  return -1;
}

int atag_parse_str(atag_container_t *container, char *option)
{
  char *p=option;
  uint32_t tag;
  
  if(atag_parse_tag(&p, &tag)<0)
    return -1;

  return atag_add_string_tag(container, tag, p);
}

int atag_parse_u32(atag_container_t *container, char *option)
{
  char *p=option;
  uint32_t tag;
  uint32_t *a, *b;
  long l;
  int i;
  
  if(atag_parse_tag(&p, &tag)<0)
    return -1;

  a=(uint32_t*)atag_add_head(container, 2, tag);
  if(!a)
    return -1;

  i=0;
  do{
   if(!*p){
     a[-2]+=i;
     return 1;
   }
   if(si_long(&p,&l,0)<0)
     break;
   b=(uint32_t*)atag_add_space(container, 4);
   *b=l;
   i++;
   if(*p==',')
     p++;
  }while(1);

  return -1;
}

int atag_container_flg;
atag_container_t atag_container;

/*----------------------------------------------------------------*/

static void
usage(void)
{
  printf("usage: flashmsc <parameters> <send_file>\n");
  printf("  -H, --hwlink <name>      select target connection HW type [%s]\n",mx1_hwlink);
  printf("  -d, --sdev <name>        name of RS232 device [%s]\n",mx1_sdev);
  printf("  -B, --baud <num>         RS232 baudrate [%d]\n",mx1_baud);
  printf("  -m, --mem-dump           dump memory range defined by -s -l -t\n");
  printf("  -s, --start  <addr>      start address of transfer\n");
  printf("  -l, --length <num>       length of upload block\n");
  printf("  -O, --mem-offs <num>     address offset\n");
  printf("  -t, --mem-type <num>     select memory type\n");
  printf("  -g, --go     <addr>      start program from address\n");
  printf("  -G, --go-entry           start program from prog. entry\n");
  printf("  -V, --version            show version\n");
  printf("  -h, --help               this usage screen\n");
  printf("  -E, --mass-erase <mode>  full or mode controlled device erase\n");
  printf("  -e, --erase		     erase region defined by -s -l\n");
  printf("  -F, --fill <num>	     fill region defined by -s -l by value\n");
  printf("  -f, --format <format>    format of data file\n");
  printf("  -P, --prop <name>=<val>  extended properties\n");
  printf("  -A, --atag-str <id>=<s>  add ATAG string\n");
  printf("  -a, --atag-u32 <id>=<a>  add ATAG array of longs\n");
  printf("  -S, --atag-base <addr>   base address of ATAG list\n");

/*
  printf("  -X, --xtal-hz <num>      target crystal frequency\n");
  printf("  -c, --command <num>      numeric command value\n");
  printf("  -b, --blockmode <num>    block size\n");
  printf("  -w, --wait-reply <num>   time to wait for reply in ms\n");
  printf("  -r, --reset              reset before download\n");
  printf("  -b, --break              send communication break character\n");
  printf("  -u, --upload             upload memory block [download]\n");
  printf("  -f, --format <format>    format of data file [intelhex]\n");
*/
}

int main(int argc, char **argv) 
{
  int res;
  /*char ch;*/
  uniload_alg_info_t algi_buff;
  uniload_alg_info_t *algi=&algi_buff;
  
  static struct option long_opts[] = {
    { "hwlink",   1, 0, 'H' },
    { "sdev",     1, 0, 'd' },
    { "udev",     1, 0, 'd' },
    { "baud",     1, 0, 'B' },
    { "command",  1, 0, 'c' },
    { "blockmode",1, 0, 'b' },
    { "wait-reply",1,0, 'w' },
    { "mass-erase",1,0, 'E' },
    { "erase",    0, 0, 'e' },
    { "fill",     1, 0, 'F' },
    { "mem-dump", 0, 0, 'm' },
    { "mem-offs", 1, 0, 'O' },
    { "mem-type", 1, 0, 't' },
    { "start",    1, 0, 's' },
    { "length",   1, 0, 'l' },
    { "go",       1, 0, 'g' },
    { "go-entry", 1, 0, 'G' },
    { "reset",    0, 0, 'r' },
    { "break",    0, 0, 'k' },
    { "upload",   0, 0, 'u' },
    { "format",   1, 0, 'f' },
    { "prop",     1, 0, 'P' },
    { "atag-str", 1, 0, 'A' },
    { "atag-u32", 1, 0, 'a' },
    { "atag-base",1, 0, 'S' },
    { "version",  0, 0, 'V' },
    { "help",     0, 0, 'h' },
    { "xtal-hz",  1, 0, 'X' },
    { 0, 0, 0, 0}
  };
  int opt;

  memset(algi,0,sizeof(*algi));

 #ifndef HAS_GETOPT_LONG
  while ((opt = getopt(argc, argv, "H:d:B:c:b:w:E:eF:mt:s:l:O:g:Grkuf:P:A:a:S:VhD:X:")) != EOF)
 #else
  while ((opt = getopt_long(argc, argv, "H:d:B:c:b:w:E:eF:mt:s:l:O:g:Grkuf:P:A:a:S:VhD:X:",
			    &long_opts[0], NULL)) != EOF)
 #endif
  switch (opt) {
    case 'H':
      mx1_hwlink=optarg;
      break;
    case 'd':
      mx1_sdev=optarg;
      break;
    case 'B':
      mx1_baud = strtol(optarg,NULL,0);
      break;
    case 'X':
      mx1_xhz = strtol(optarg,NULL,0);
      break;
    case 'c':
      if(optarg[0]=='B'){
        command=0;
	break;
      }
      command = strtol(optarg,NULL,0);
      break;
    case 'b':
      blockmode = strtol(optarg,NULL,0);
      break;
    case 'w':
      mx1_waitrep = strtol(optarg,NULL,0)*1000;
      break;
    case 'E':
      masserase_end = strtol(optarg,NULL,0);
      masserase_flg=1;
      break;
    case 'e':
      regerase_flg=1;
      break;
    case 'F':
      fill_flg=1;
      if(add_to_arr((void**)&fill_pat_val,&fill_pat_len,0,optarg)<0){
	fprintf(stderr,"%s: incorrect patern data \"%s\"\n",argv[0],optarg);
	exit(2);
      }
      if(!fill_pat_len){
	fprintf(stderr,"%s: incorrect patern data - empty value\n",argv[0]);
	exit(2);
      }
      break;
    case 'm':
      dump_flg=1;
      break;
    case 't':
      mem_type=strtoul(optarg,NULL,0);
      break;
    case 's':
      mem_start=strtoul(optarg,NULL,0);
      break;
    case 'l':
      mem_length=strtoul(optarg,NULL,0);
      break;
    case 'O':
      mem_offs=strtoul(optarg,NULL,0);
      break;
    case 'g':
      go_addr=strtoul(optarg,NULL,0);
    case 'G':
      go_flg=1;
      break;
    case 'r':
      reset_flg=1;
      break;
    case 'k':
      break_flg=1;
      break;
    case 'u':
      upload_flg=1;
      break;
    case 'f':
      file_format=optarg;
      break;
    case 'P':
      if(extprop_parse2list(&extprop_list, optarg)<0){
	fprintf(stderr,"%s: incorrect extended property %s\n",argv[0],optarg);
	exit(2);
      }
      break;
    case 'A':
      if(atag_parse_str(&atag_container, optarg)<0){
	fprintf(stderr,"%s: incorrect atag string %s\n",argv[0],optarg);
	exit(2);
      }
      break;
    case 'a':
      if(atag_parse_u32(&atag_container, optarg)<0){
	fprintf(stderr,"%s: incorrect atag u32 %s\n",argv[0],optarg);
	exit(2);
      }
      break;
    case 'S':
      atag_container_flg=1;
      atag_container.base=strtoul(optarg,NULL,0);
      break;
    case 'V':
      fputs("tohit pre alpha\n", stdout);
      exit(0);
    case 'h':
    default:
      usage();
      exit(opt == 'h' ? 0 : 1);
  }

  if(atag_container_flg){
    atag_add_space(&atag_container, 8);
  }

  if(!strcmp(mx1_hwlink,"tagimage")){
    res=mx1_tagimage_open(algi, mx1_sdev, 0);
    if(res<0){
      fprintf(stderr,"TAGIMAGE open failed\n");
      return -1;
    }
  }else if(!strcmp(mx1_hwlink,"usb")){
    res=mx1_usb_open(algi, mx1_sdev);
    if(res<0){
      fprintf(stderr,"USB open failed\n");
      return -1;
    }
  }else{
    res=mx1_rs232_open(algi, mx1_sdev, mx1_baud, 0, 0);
    if(res<0){
      fprintf(stderr,"RS-232 open failed\n");
      return -1;
    }
  }
  algi->extprop=extprop_list;
  algi->def_mem_type=mem_type;
  algi->mem_offs=mem_offs;
  if(!algi->max_blksize){
    algi->max_blksize=512;
  }

  /*mx1_xhz*/
  if(uniload_setup(algi)<0){
    fprintf(stderr,"Target setup failed\n");
    close(algi->dev_fd);
    return -1;
  }

  sleep(1);
  tcflush(algi->dev_fd, TCIOFLUSH);

  if(reset_flg){
  }

  if(masserase_flg){
    if(uniload_masserase(algi,masserase_end)<0){
      fprintf(stderr,"Mass Erase failed\n");
      close(mx1_fd);
      return -1;
    }
  }

  if(regerase_flg){
    res=uniload_regerase(algi, mem_start, mem_length);
    if(res<0){
      fprintf(stderr,"mx1_regerase: returned %d\n",res);
      exit(1);
    }
  }

  if(fill_flg) {
    uniload_mem_fill(algi, mem_type, mem_start, mem_length?mem_length:fill_pat_len, fill_pat_val, fill_pat_len);
  }

  if(!upload_flg){
    while(optind<argc) {
      uniload_load_use_lma=1;
      if(uniload_download_file(algi,argv[optind],file_format,NULL,&entry_addr)<0){
	fprintf(stderr,"Error to download file %s\n", argv[optind]);
	exit(1); 
      }
      optind++;
    }
  }else if(upload_flg&&(optind<argc)){
    if(optind+1!=argc) 
    { printf("upload_file : needs exactly one filename\n");
      exit(1);
    };
    if(uniload_upload_file(algi, argv[optind], file_format, 
                           mem_type, mem_start, mem_length)<0){
      exit(2);
    }
  } 

  if(dump_flg){
    res=uniload_mem_dump(algi, mem_type, mem_start, mem_length);
    if(res<0){
      fprintf(stderr,"uniload_mem_dump: returned %d\n",res);
      exit(1);
    }
  }

  if(atag_container_flg){
    res=uniload_mem_write(algi, mem_type, atag_container.base, atag_container.size, atag_container.data);
    if(res<0){
      fprintf(stderr,"uniload_mem_write: returned %d for ATAG list\n",res);
      exit(1);
    }
  }

  if(go_flg){
    if(go_addr)
      uniload_go_addr(algi, go_addr);
    else
      uniload_go_addr(algi, entry_addr);
  }
  
  uniload_release(algi);
  
  return 0;
}
